/* ------------------------------------------------------------------------- */
/*
 *  hcss.h
 *
 *  Copyright (c) 2007 "clown". All rights reserved.
 *
 *  Permission to copy, use, modify and distribute this software is
 *  granted provided this copyright notice appears in all copies.
 *  This software is provided "as is" without express or implied
 *  warranty, and with no claim as to its suitability for any purpose.
 *
 *  Last-modified: Thu 05 Apr 2007 23:09:32 JST
 */
/* ------------------------------------------------------------------------- */
#include <iostream>
#include <fstream>
#include <sstream>
#include <string>
#include <vector>
#include <iterator>
#include <stdexcept>

#ifndef CLX_HCSS_H
#define CLX_HCSS_H

namespace clx {
	/* --------------------------------------------------------------------- */
	//  basic_hierarchical_css
	/* --------------------------------------------------------------------- */
	template <
		class CharT,
		class Traits = std::char_traits<CharT>
	>
	class basic_hierarchical_css {
	public:
		typedef CharT char_type;
		typedef unsigned int size_type;
		typedef std::basic_string<CharT, Traits> string_type;
		typedef std::vector<string_type> selector_list;
		typedef std::vector<string_type> declaration_list;
		
		basic_hierarchical_css() : tags_(), v_(), sp_(), comments_() {}
		
		template <class InputIterator>
		explicit basic_hierarchical_css(InputIterator first, InputIterator last) :
			tags_(), v_(), sp_(), comments_() {
			this->assign(first, last);
		}
		
		explicit basic_hierarchical_css(const char_type* filename) :
			tags_(), v_(), sp_(), comments_() {
			this->assign(filename);
		}
		
		explicit basic_hierarchical_css(const string_type& filename) :
			tags_(), v_(), sp_(), comments_() {
			this->assign(filename);
		}
		
		// operator
		template <class InputIterator>
		basic_hierarchical_css& assign(InputIterator first, InputIterator last) {
			string_type tmp;
			while (first != last) {
				if (this->is_comment(*first)) {
					first++;
					continue;
				}
				
				tmp += *first++;
				if (tmp.find('{') != string_type::npos) this->push_entry(tmp);
				else if (tmp.find(';') != string_type::npos) this->push_property(tmp);
				else if (tmp.find('}') != string_type::npos) {
					if (tags_.empty()) throw std::runtime_error("too many \"}\" symbol");
					this->pop_tag(tmp);
				}
				else tmp += ' ';
			}
			
			// slight syntax checking
			if (comments_ > 0) throw std::runtime_error("too few \"*/\" symbol");
			if (!tags_.empty()) throw std::runtime_error("too few \"}\" symbol");
			
			return *this;
		}
		
		basic_hierarchical_css& assign(const char_type* filename) {
			std::basic_ifstream<CharT, Traits> fs(filename);
			if (fs.fail()) {
				std::basic_stringstream<CharT, Traits> msg;
				msg << filename << ": No such file or directory";
				throw std::runtime_error(msg.str());
			}
			
			std::istream_iterator<string_type> first(fs);
			std::istream_iterator<string_type> last;
			this->assign(first, last);
			
			fs.close();
			return *this;
		}
		
		basic_hierarchical_css& assign(const string_type& filename) {
			return this->assign(filename.c_str());
		}
		
		// accessor
		size_type size() const { return v_.size(); }
		const string_type& tag(size_type i) const { return v_.at(i).first; }
		const declaration_list& property(size_type i) const { return v_.at(i).second; }
		
	private:
		typedef std::pair<string_type, declaration_list> decl_block;
		typedef std::vector<decl_block> container;
		
		selector_list tags_;
		container v_;
		
		std::vector<size_type> sp_;
		size_type comments_;
		
		
		// invalid copy constructor and self-assign operator
		explicit basic_hierarchical_css(const basic_hierarchical_css& cp) {}
		
		basic_hierarchical_css& operator=(const basic_hierarchical_css& cp) {
			return *this;
		}
		
		bool is_comment(const string_type& s) {
			if (s.find("/*") != string_type::npos) {
				comments_++;
				return true;
			}
			else if (s.find("*/") != string_type::npos) {
				if (comments_ == 0) throw std::runtime_error("too many \"*/\" symbol");
				comments_--;
				return true;
			}
			else if (comments_ > 0) return true;
			return false;
		}
		
		void push_entry(string_type& s) {
			s.erase(s.find('{'));
			if (s.find_last_of(' ') != string_type::npos) s.erase(s.find_last_of(' '));
			tags_.push_back(s);
			
			std::basic_stringstream<CharT, Traits> ss;
			for (size_type i = 0; i < tags_.size(); i++) {
				if (tags_.at(i) != "*" && tags_.at(i) != "body") {
					ss << tags_.at(i) << ' ';
				}
			}
			
			decl_block tmp;
			if (ss.str().empty()) tmp.first = tags_.at(tags_.size() - 1);
			else tmp.first = ss.str().substr(0, ss.str().find_last_of(' '));
			v_.push_back(tmp);
			sp_.push_back(v_.size() - 1);
			
			s.erase();
		}
		
		void push_property(string_type& s) {
			v_.at(sp_.back()).second.push_back(s);
			s.erase();
		}
		void pop_tag(string_type& s) {
			tags_.pop_back();
			sp_.pop_back();
			s.erase();
		}
	};
	
	typedef basic_hierarchical_css<char> hcss;
}

#endif // CLX_HCSS_H
